/************************************************************
 *
 *  OTCronItem.hpp
 *
 */

/************************************************************
 -----BEGIN PGP SIGNED MESSAGE-----
 Hash: SHA1

 *                 OPEN TRANSACTIONS
 *
 *       Financial Cryptography and Digital Cash
 *       Library, Protocol, API, Server, CLI, GUI
 *
 *       -- Anonymous Numbered Accounts.
 *       -- Untraceable Digital Cash.
 *       -- Triple-Signed Receipts.
 *       -- Cheques, Vouchers, Transfers, Inboxes.
 *       -- Basket Currencies, Markets, Payment Plans.
 *       -- Signed, XML, Ricardian-style Contracts.
 *       -- Scripted smart contracts.
 *
 *  Copyright (C) 2010-2013 by "Fellow Traveler" (A pseudonym)
 *
 *  EMAIL:
 *  FellowTraveler@rayservers.net
 *
 *  BITCOIN:  1NtTPVVjDsUfDWybS4BwvHpG2pdS9RnYyQ
 *
 *  KEY FINGERPRINT (PGP Key in license file):
 *  9DD5 90EB 9292 4B48 0484  7910 0308 00ED F951 BB8E
 *
 *  OFFICIAL PROJECT WIKI(s):
 *  https://github.com/FellowTraveler/Moneychanger
 *  https://github.com/FellowTraveler/Open-Transactions/wiki
 *
 *  WEBSITE:
 *  http://www.OpenTransactions.org/
 *
 *  Components and licensing:
 *   -- Moneychanger..A Java client GUI.....LICENSE:.....GPLv3
 *   -- otlib.........A class library.......LICENSE:...LAGPLv3
 *   -- otapi.........A client API..........LICENSE:...LAGPLv3
 *   -- opentxs/ot....Command-line client...LICENSE:...LAGPLv3
 *   -- otserver......Server Application....LICENSE:....AGPLv3
 *  Github.com/FellowTraveler/Open-Transactions/wiki/Components
 *
 *  All of the above OT components were designed and written by
 *  Fellow Traveler, with the exception of Moneychanger, which
 *  was contracted out to Vicky C (bitcointrader4@gmail.com).
 *  The open-source community has since actively contributed.
 *
 *  -----------------------------------------------------
 *
 *   LICENSE:
 *   This program is free software: you can redistribute it
 *   and/or modify it under the terms of the GNU Affero
 *   General Public License as published by the Free Software
 *   Foundation, either version 3 of the License, or (at your
 *   option) any later version.
 *
 *   ADDITIONAL PERMISSION under the GNU Affero GPL version 3
 *   section 7: (This paragraph applies only to the LAGPLv3
 *   components listed above.) If you modify this Program, or
 *   any covered work, by linking or combining it with other
 *   code, such other code is not for that reason alone subject
 *   to any of the requirements of the GNU Affero GPL version 3.
 *   (==> This means if you are only using the OT API, then you
 *   don't have to open-source your code--only your changes to
 *   Open-Transactions itself must be open source. Similar to
 *   LGPLv3, except it applies to software-as-a-service, not
 *   just to distributing binaries.)
 *
 *   Extra WAIVER for OpenSSL, Lucre, and all other libraries
 *   used by Open Transactions: This program is released under
 *   the AGPL with the additional exemption that compiling,
 *   linking, and/or using OpenSSL is allowed. The same is true
 *   for any other open source libraries included in this
 *   project: complete waiver from the AGPL is hereby granted to
 *   compile, link, and/or use them with Open-Transactions,
 *   according to their own terms, as long as the rest of the
 *   Open-Transactions terms remain respected, with regard to
 *   the Open-Transactions code itself.
 *
 *   Lucre License:
 *   This code is also "dual-license", meaning that Ben Lau-
 *   rie's license must also be included and respected, since
 *   the code for Lucre is also included with Open Transactions.
 *   See Open-Transactions/src/otlib/lucre/LUCRE_LICENSE.txt
 *   The Laurie requirements are light, but if there is any
 *   problem with his license, simply remove the Lucre code.
 *   Although there are no other blind token algorithms in Open
 *   Transactions (yet. credlib is coming), the other functions
 *   will continue to operate.
 *   See Lucre on Github:  https://github.com/benlaurie/lucre
 *   -----------------------------------------------------
 *   You should have received a copy of the GNU Affero General
 *   Public License along with this program.  If not, see:
 *   http://www.gnu.org/licenses/
 *
 *   If you would like to use this software outside of the free
 *   software license, please contact FellowTraveler.
 *   (Unfortunately many will run anonymously and untraceably,
 *   so who could really stop them?)
 *
 *   DISCLAIMER:
 *   This program is distributed in the hope that it will be
 *   useful, but WITHOUT ANY WARRANTY; without even the implied
 *   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *   PURPOSE.  See the GNU Affero General Public License for
 *   more details.

 -----BEGIN PGP SIGNATURE-----
 Version: GnuPG v1.4.9 (Darwin)

 iQIcBAEBAgAGBQJRSsfJAAoJEAMIAO35UbuOQT8P/RJbka8etf7wbxdHQNAY+2cC
 vDf8J3X8VI+pwMqv6wgTVy17venMZJa4I4ikXD/MRyWV1XbTG0mBXk/7AZk7Rexk
 KTvL/U1kWiez6+8XXLye+k2JNM6v7eej8xMrqEcO0ZArh/DsLoIn1y8p8qjBI7+m
 aE7lhstDiD0z8mwRRLKFLN2IH5rAFaZZUvj5ERJaoYUKdn4c+RcQVei2YOl4T0FU
 LWND3YLoH8naqJXkaOKEN4UfJINCwxhe5Ke9wyfLWLUO7NamRkWD2T7CJ0xocnD1
 sjAzlVGNgaFDRflfIF4QhBx1Ddl6wwhJfw+d08bjqblSq8aXDkmFA7HeunSFKkdn
 oIEOEgyj+veuOMRJC5pnBJ9vV+7qRdDKQWaCKotynt4sWJDGQ9kWGWm74SsNaduN
 TPMyr9kNmGsfR69Q2Zq/FLcLX/j8ESxU+HYUB4vaARw2xEOu2xwDDv6jt0j3Vqsg
 x7rWv4S/Eh18FDNDkVRChiNoOIilLYLL6c38uMf1pnItBuxP3uhgY6COm59kVaRh
 nyGTYCDYD2TK+fI9o89F1297uDCwEJ62U0Q7iTDp5QuXCoxkPfv8/kX6lS6T3y9G
 M9mqIoLbIQ1EDntFv7/t6fUTS2+46uCrdZWbQ5RjYXdrzjij02nDmJAm2BngnZvd
 kamH0Y/n11lCvo1oQxM+
 =uSzz
 -----END PGP SIGNATURE-----
 **************************************************************/

// Base class for OTTrade and OTAgreement.
// OTCron contains lists of these for regular processing.

#ifndef OPENTXS_CORE_CRON_OTCRONITEM_HPP
#define OPENTXS_CORE_CRON_OTCRONITEM_HPP

#include <opentxs/core/OTTrackable.hpp>

#include <deque>

namespace opentxs
{

class OTCron;
class OTIdentifier;
class OTPseudonym;
class OTString;

class OTCronItem : public OTTrackable
{
public:
    OTCronItem();

private: // Private prevents erroneous use by other classes.
    typedef OTTrackable ot_super;

private:
    OTCron* m_pCron;
    OTPseudonym* serverNym_;
    OTIdentifier* serverId_;
    time64_t m_CREATION_DATE;     // The date, in seconds, when the CronItem was
                                  // authorized.
    time64_t m_LAST_PROCESS_DATE; // The last time this item was processed.
    int64_t m_PROCESS_INTERVAL;   // How often to Process Cron on this item.

protected:
    std::deque<int64_t> m_dequeClosingNumbers; // Numbers used for CLOSING a
                                               // transaction. (finalReceipt.)

protected:
    OTCronItem(const OTIdentifier& SERVER_ID, const OTIdentifier& ASSET_ID);
    OTCronItem(const OTIdentifier& SERVER_ID, const OTIdentifier& ASSET_ID,
               const OTIdentifier& ACCT_ID, const OTIdentifier& USER_ID);

    OTIdentifier* m_pCancelerNymID;

    bool m_bCanceled; // This defaults to false. But if someone cancels it
                      // (BEFORE it is ever activated, just to nip it in the bud
                      // and harvest the numbers, and send the notices, etc) --
                      // then we set this to true, and we also set the canceler
                      // Nym ID. (So we can see these values later and know
                      // whether it was canceled before activation, and if so,
                      // who did it.)

    bool m_bRemovalFlag; // Set this to true and the cronitem will be removed
                         // from Cron on next process.
    // (And its offer will be removed from the Market as well, if appropriate.)
    virtual void onActivate()
    {
    } // called by HookActivationOnCron().

    virtual void onFinalReceipt(
        OTCronItem& theOrigCronItem, const int64_t& lNewTransactionNumber,
        OTPseudonym& theOriginator,
        OTPseudonym* pRemover); // called by HookRemovalFromCron().

    virtual void onRemovalFromCron()
    {
    } // called by HookRemovalFromCron().
    void ClearClosingNumbers();

public:
    // To force the Nym to close out the closing number on the receipt.
    bool DropFinalReceiptToInbox(
        const OTIdentifier& USER_ID, const OTIdentifier& ACCOUNT_ID,
        const int64_t& lNewTransactionNumber, const int64_t& lClosingNumber,
        const OTString& strOrigCronItem, OTString* pstrNote = nullptr,
        OTString* pstrAttachment = nullptr, OTAccount* pActualAcct = nullptr);

    // Notify the Nym that the OPENING number is now closed, so he can remove it
    // from his issued list.
    bool DropFinalReceiptToNymbox(const OTIdentifier& USER_ID,
                                  const int64_t& lNewTransactionNumber,
                                  const OTString& strOrigCronItem,
                                  OTString* pstrNote = nullptr,
                                  OTString* pstrAttachment = nullptr,
                                  OTPseudonym* pActualNym = nullptr);
    virtual bool CanRemoveItemFromCron(OTPseudonym& theNym);
    virtual void HarvestOpeningNumber(OTPseudonym& theNym);
    virtual void HarvestClosingNumbers(OTPseudonym& theNym);
    // pActivator and pRemover are both "SOMETIMES nullptr"
    // I don't default the parameter, because I want to force the programmer to
    // choose.

    // Called in OTCron::AddCronItem.
    void HookActivationOnCron(OTPseudonym* pActivator,
                              bool bForTheFirstTime = false); // This calls
                                                              // onActivate,
                                                              // which is
                                                              // virtual.

    // Called in OTCron::RemoveCronItem as well as OTCron::ProcessCron.
    // This calls onFinalReceipt, then onRemovalFromCron. Both are virtual.
    void HookRemovalFromCron(OTPseudonym* pRemover, int64_t newTransactionNo);

    inline bool IsFlaggedForRemoval() const
    {
        return m_bRemovalFlag;
    }
    inline void FlagForRemoval()
    {
        m_bRemovalFlag = true;
    }
    inline void SetCronPointer(OTCron& theCron)
    {
        m_pCron = &theCron;
    }

    EXPORT static OTCronItem* NewCronItem(const OTString& strCronItem);
    EXPORT static OTCronItem* LoadCronReceipt(
        const int64_t& lTransactionNum); // Server-side only.
    EXPORT static OTCronItem* LoadActiveCronReceipt(
        const int64_t& lTransactionNum,
        const OTIdentifier& serverID); // Client-side only.
    EXPORT static bool EraseActiveCronReceipt(
        const int64_t& lTransactionNum, const OTIdentifier& nymID,
        const OTIdentifier& serverID); // Client-side only.
    EXPORT static bool GetActiveCronTransNums(OTNumList& output, // Client-side
                                                                 // only.
                                              const OTIdentifier& nymID,
                                              const OTIdentifier& serverID);
    inline void SetCreationDate(const time64_t& CREATION_DATE)
    {
        m_CREATION_DATE = CREATION_DATE;
    }
    inline const time64_t& GetCreationDate() const
    {
        return m_CREATION_DATE;
    }

    EXPORT bool SetDateRange(time64_t VALID_FROM = OT_TIME_ZERO,
                             time64_t VALID_TO = OT_TIME_ZERO);
    inline void SetLastProcessDate(const time64_t& THE_DATE)
    {
        m_LAST_PROCESS_DATE = THE_DATE;
    }
    inline const time64_t& GetLastProcessDate() const
    {
        return m_LAST_PROCESS_DATE;
    }

    inline void SetProcessInterval(const int64_t& THE_DATE)
    {
        m_PROCESS_INTERVAL = THE_DATE;
    }
    inline const int64_t& GetProcessInterval() const
    {
        return m_PROCESS_INTERVAL;
    }

    inline OTCron* GetCron() const
    {
        return m_pCron;
    }
    void setServerNym(OTPseudonym* serverNym)
    {
        serverNym_ = serverNym;
    }
    void setServerId(OTIdentifier* serverId)
    {
        serverId_ = serverId;
    }
    // When first adding anything to Cron, a copy needs to be saved in a folder
    // somewhere.
    EXPORT bool SaveCronReceipt(); // server side only
    EXPORT bool SaveActiveCronReceipt(const OTIdentifier& theNymID); // client
                                                                     // side
                                                                     // only

    // Return True if should stay on OTCron's list for more processing.
    // Return False if expired or otherwise should be removed.
    virtual bool ProcessCron(); // OTCron calls this regularly, which is my
                                // chance to expire, etc.
                                // From OTTrackable (parent class of this)
    virtual ~OTCronItem();

    void InitCronItem();

    virtual void Release();
    void Release_CronItem();
    EXPORT bool GetCancelerID(OTIdentifier& theOutput) const;
    EXPORT bool IsCanceled() const
    {
        return m_bCanceled;
    }
    EXPORT bool CancelBeforeActivation(
        OTPseudonym& theCancelerNym); // When canceling a cron item before it
                                      // has been activated, use this.
    // These are for     std::deque<int64_t> m_dequeClosingNumbers;
    // They are numbers used for CLOSING a transaction. (finalReceipt.)

    EXPORT int64_t GetClosingTransactionNoAt(uint32_t nIndex) const;
    EXPORT int32_t GetCountClosingNumbers() const;

    EXPORT void AddClosingTransactionNo(const int64_t& lClosingTransactionNo);

    // HIGHER LEVEL ABSTRACTIONS:
    EXPORT int64_t GetOpeningNum() const;
    EXPORT int64_t GetClosingNum() const;
    virtual bool IsValidOpeningNumber(const int64_t& lOpeningNum) const;

    virtual int64_t GetOpeningNumber(const OTIdentifier& theNymID) const;
    virtual int64_t GetClosingNumber(const OTIdentifier& theAcctID) const;
    virtual int32_t ProcessXMLNode(irr::io::IrrXMLReader*& xml);
    //    virtual void UpdateContents(); // Before transmission or
    // serialization, this is where the ledger saves its contents
    //    virtual bool SaveContractWallet(std::ofstream& ofs) const;
};

} // namespace opentxs

#endif // OPENTXS_CORE_CRON_OTCRONITEM_HPP
